/* ============================================================
 *
 * This file is a part of digiKam project
 * https://www.digikam.org
 *
 * Date        : 2010-11-10
 * Description : meta-filter to apply FilterActions
 *
 * SPDX-FileCopyrightText: 2010-2011 by Marcel Wiesweg <marcel dot wiesweg at gmx dot de>
 * SPDX-FileCopyrightText: 2010      by Martin Klapetek <martin dot klapetek at gmail dot com>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * ============================================================ */

#pragma once

// Local includes

#include "digikam_export.h"
#include "dimgthreadedfilter.h"
#include "filteraction.h"

namespace Digikam
{

class DIGIKAM_EXPORT FilterActionFilter : public DImgThreadedFilter
{
    Q_OBJECT

public:

    /**
     * @brief A meta-filter applying other filter according to a list of FilterActions
     */
    explicit FilterActionFilter(QObject* const parent = nullptr);
    ~FilterActionFilter()                             override;

    /**
     * @brief Per default, the filter will stop when it encounters an unsupported action.
     * If you want it to continue, set this to true.
     * Only the last occurred error will then be reported.
     */
    void setContinueOnError(bool cont);

    /**
     * @brief Set or add to existing list of the given filter actions
     */
    void setFilterActions(const QList<FilterAction>& actions);
    void addFilterActions(const QList<FilterAction>& actions);
    void setFilterAction(const FilterAction& action);
    void addFilterAction(const FilterAction& action);

    QList<FilterAction> filterActions()         const;

    /**
     * @return the list of applied filter actions. This is probably identical
     * to filterActions, but it can differ in some situations:
     *
     * - If completelyApplied() is false, it will contain only the successful actions
     * - The list is regenerated by the filters. If filterActions contains
     *   actions with an older version, still supported by the filter,
     *   the filter will now possibly return the newer, current version
     */
    QList<FilterAction> appliedFilterActions()  const;

    /**
     * @return True if all FilterActions are reproducible
     */
    bool isReproducible()                       const;

    /**
     * @return True if all FilterActions are reproducible
     * or are ComplexFilters. That means the identical result
     * may not be reproducible, but a sufficiently similar result
     * may be available and apply will probably complete.
     */
    bool isComplexAction()                      const;

    /**
     * @return true if all actions are supported.
     */
    bool isSupported()                          const;

    /**
     * @brief After the thread was run, you can find out if application was successful.
     * A precondition is that at least isComplexAction() and isSupported() returns true.
     * If all filters applied cleanly, completelyApplied() returns true.
     * appliedActions() returns all applied actions, if completelyApplied(),
     * the same as filterActions().
     * If not completelyApplied, failedAction() returns the action that failed,
     * failedActionIndex its index in filterActions(), and failedActionMessage
     * an optional error message.
     *
     * @note The finished state (true) does not mean that completelyApplied() is also true.
     */
    bool         completelyApplied()            const;
    FilterAction failedAction()                 const;
    int          failedActionIndex()            const;
    QString      failedActionMessage()          const;

    /**
     * @brief These methods do not make sense here. Use filterActions.
     */
    FilterAction filterAction()                       override
    {
        return FilterAction();
    }

    void readParameters(const FilterAction&)          override
    {
    }

    QString filterIdentifier()                  const override
    {
        return QString();
    }

protected:

    void filterImage()                                override;

private:

    class Private;
    Private* const d = nullptr;
};

} // namespace Digikam
